#include <libtcod.hpp>
#include <stdio.h>
#include "build.h"
#include "map.h"
#include "vec2.h"
#include "atlas.h"
#include "buf.h"
#include "text.h"

class RCT
{
public:
    int		x, y, w, h;

    bool	overlap(const RCT &o) const
    {
	if (MAX(x, o.x) > MIN(x+w, o.x+o.w))
	    return false;
	if (MAX(y, o.y) > MIN(y+h, o.y+o.h))
	    return false;
	return true;
    }

    int		midx() const { return x + w / 2; }
    int		midy() const { return y + h / 2; }
};

class bu_vtx
{
public:
    bu_vtx() : pos(0, 0) { rad = 0; }
    bu_vtx(float ix, float iy) : pos(ix, iy) { rad = 0.0; }

    VEC2		pos;
    float		rad;
};

struct bu_edge
{
public:
    bu_edge() { vtx[0] = -1; vtx[1] = -1; }
    bu_edge(int a, int b) { vtx[0] = a; vtx[1] = b; }

    float		dist(VEC2 p, const PTRLIST<bu_vtx> &vtxlist)
    {
	VEC2		proj, linedir;
	proj = p;
	proj -= vtxlist(vtx[0]).pos;
	linedir = vtxlist(vtx[1]).pos;
	linedir -= vtxlist(vtx[0]).pos;

	float t = dot(proj, linedir) / linedir.quadrature();

	t = BOUND(t, 0, 1);
	float radius = lerp(t, vtxlist(vtx[0]).rad,
			       vtxlist(vtx[1]).rad);

	proj = vtxlist(vtx[0]).pos;
	proj += t * linedir;

	return ::dist(proj, p) - radius;
    }

    int			vtx[2];
};

static float
computeLineDist(VEC2 p, const PTRLIST<bu_edge> &edgelist, const PTRLIST<bu_vtx> &vtxlist)
{
    float	dist = 1e10f;
    for (int i = 0; i < edgelist.entries(); i++)
    {
	dist = MIN(dist, edgelist(i).dist(p, vtxlist));
    }
    return dist;
}

static void
createSplat(ROOMDATA<float> &data,
	    VEC2 splatcenter,
	    VEC2 splatsize,
	    int nlines,
	    float centerrad,
	    bool buildfingers, float fingerminrad, float fingermaxrad,
	    float noisesize, float noisedisplace)
{
    PTRLIST<bu_vtx>	 vtxlist;
    PTRLIST<bu_edge>	 edgelist;

    // First vertex is at 0,0
    vtxlist.append(bu_vtx(0, 0));

    for (int i = 0; i < nlines; i++)
    {
	// pick a random edge.
	int		startvtx = rand_choice(vtxlist.entries());
	bu_vtx		endpos;

	endpos.pos = vtxlist(startvtx).pos + VEC2::randUnit() * rand_frange(0.5, 1.5);
	int		endvtx = vtxlist.append(endpos);
	edgelist.append(bu_edge(startvtx, endvtx));
    }

    // Compute, first pass we compute our index count.
    for (int i = 0; i < edgelist.entries(); i++)
    {
	vtxlist(edgelist(i).vtx[0]).rad += 1.0;
	vtxlist(edgelist(i).vtx[1]).rad += 1.0;
    }

    // Now remap...
    for (int i = 0; i < vtxlist.entries(); i++)
    {
	vtxlist(i).rad -= 1;
	if (vtxlist(i).rad > 0.5)
	    vtxlist(i).rad = centerrad;
    }

    // Expand all fingers.
    if (buildfingers)
    {
	int		maxvtx = vtxlist.entries();
	for (int i = 0; i < maxvtx; i++)
	{
	    if (vtxlist(i).rad < 0.001)
	    {
		// pick a random edge.
		int		startvtx = i;
		bu_vtx	endpos;

		endpos.pos = vtxlist(startvtx).pos + VEC2::randUnit() * rand_frange(0.5, 1.5);
		endpos.rad = fingerminrad;
		int		endvtx = vtxlist.append(endpos);
		bu_vtx	startpos;
		startpos.pos = vtxlist(startvtx).pos;
		startpos.rad = fingermaxrad;
		startvtx = vtxlist.append(startpos);

		edgelist.append(bu_edge(startvtx, endvtx));
	    }
	}
    }

    VEC2		mincoord, maxcoord;

    mincoord = maxcoord = vtxlist(0).pos;
    for (int i = 1; i < vtxlist.entries(); i++)
    {
	mincoord = min(mincoord, vtxlist(i).pos);
	maxcoord = max(maxcoord, vtxlist(i).pos);
    }

    // Recenter & rescale to 10 .. 90
    float		size = (maxcoord - mincoord).maxcomp();
    VEC2		centeradjust = (maxcoord + mincoord) * 0.5;

    for (int i = 0; i < vtxlist.entries(); i++)
    {
	vtxlist(i).pos -= centeradjust;
	vtxlist(i).pos *= splatsize / size;
	vtxlist(i).pos += splatcenter;
    }

    // Allocate a noise map
    TCODRandom		r(rand_int());
    TCODNoise		noisex(2, &r), noisey(2, &r);

    // And compute a distance map.
    for (int y = 0; y < data.height(); y++)
    {
	for (int x = 0; x < data.width(); x++)
	{
	    VEC2		p((float)x, (float)y);

	    float		linepos[2];
	    linepos[0] = ((float)x)/noisesize;
	    linepos[1] = ((float)y)/noisesize;
	    float		dx, dy;
	    dx = noisedisplace * noisex.getTurbulenceSimplex(linepos, 3);
	    dy = noisedisplace * noisey.getTurbulenceSimplex(linepos, 3);
	    p.x() += dx;
	    p.y() += dy;

	    float		dist = computeLineDist(p, edgelist, vtxlist);
	    data(x, y) = MIN(data(x,y), dist);
	}
    }
}

ROOM_DEF *
BUILD::build(BUILD_NAME buildtype, int data, const char *name, const char *quest)
{
    ROOM_DEF_NC		*def = 0;

    
    switch (buildtype)
    {
	case BUILD_NONE:
	case NUM_BUILDS:
	    assert(0);
	    break;

	case BUILD_WORLD:
	    while (!def)
		def = buildWorld();
	    break;
	case BUILD_ROOMS:
	    while (!def)
		def = buildRoomsAndCorridors(48, 25, data);
	    break;
    }

    if (name)
	def->name = strdup(name);

    // Additional linking?
    switch (buildtype)
    {
	case BUILD_WORLD:
	    // World is self-wrapped.
	    def->west = def->east = def->name;
	    break;

	default:
	    break;
    }

    return (ROOM_DEF *) def;
}


static ITEM_NAMES
randomBook()
{
    const int BOOK_COUNT = 7;
    static const ITEM_NAMES	randsrc[BOOK_COUNT] =
    {
	ITEM_BOOK_SHADOW,
	ITEM_BOOK_WILDGROWTH,
	ITEM_BOOK_VOLCANO,
	ITEM_BOOK_CAVEIN,
	ITEM_BOOK_RECALL,
	ITEM_BOOK_RESTORE,
	ITEM_BOOK_SHADOWSTEP
    };

    return randsrc[rand_choice(BOOK_COUNT)];
}

ROOM_DEF_NC *
BUILD::buildWorld()
{
    ROOM_DEF_NC		*def = allocRoom(48, 25);

    fill(def, SQUARE_GRASS);
    rect(def, 0, 0, 48, 10, SQUARE_SKY);

    // Build some interesting biomes
#define NUM_BIOME 6
    const SQUARE_NAMES	biome[NUM_BIOME] = 
    { 
	SQUARE_HILLS,
	SQUARE_DESERT,
	SQUARE_DESERT,
	SQUARE_FOREST,
	SQUARE_DESERT,
	SQUARE_DESERT,
    };
    for (int curbiome = 0; curbiome < NUM_BIOME; curbiome++)
    {
	int		px, py;

	if (!randomSquare(def, SQUARE_GRASS, px, py))
	{
	    delete def;
	    return 0;
	}
	ROOMDATA<float>	 biomemap(50, 50, 1000);

	createSplat(biomemap, VEC2(25, 25), VEC2(10, 10),
			/*nlines=*/8,
			/*centerrad=*/3,
			/*buildfingers=*/false,
			/*fingerminmax=*/-10, -5,
			/*noisesize=*/10,
			/*noisedisplace=*/12
			);

	// And apply the biome
	for (int y = 0; y < def->size.y; y++)
	{
	    for (int x = 0; x < def->size.x; x++)
	    {

		float		dist = biomemap.get(x+25-px, y+25-py);
		if (dist < 0)
		{
		    SQUARE_NAMES	curval;
		    curval = get(def, x, y);
		    if (curval == SQUARE_HILLS || curval == SQUARE_GRASS)
		    {
			set(def, x, y, biome[curbiome]);
		    }
		}
	    }
	}
    }

    replace(def, SQUARE_HILLS, SQUARE_FOREST, 5);
    replace(def, SQUARE_GRASS, SQUARE_FOREST, 2);

    for (int y = 0; y < def->size.y; y += 1)
    {
	for (int x = 0; x < def->size.x; x += 1)
	{
	    if (get(def, x, y) == SQUARE_FOREST)
	    {
		// Add tree.
		set(def, x, y, SQUARE_GRASS);
		addItem(def, x, y, ITEM_TREE);

	    }
	}
    }

    
    if (0)
    {
	// Debug add items.
	for (int i = 0; i < 10; i++)
	{
	    int		x, y;

	    randomSquare(def, SQUARE_GRASS, x, y);
	    addItem(def, x, y, randomBook());
	}
	for (int i = 0; i < 4; i++)
	{
	    int		x, y;

	    randomSquare(def, SQUARE_GRASS, x, y);
	    addItem(def, x, y, ITEM_DAYSTAR);
	}
    }

    // Add the altars.
    int		naltar = 4, ntries = 1000;
    while (naltar)
    {
	ntries--;
	if (!ntries)
	{
	    delete def;
	    return 0;
	}

	int		x = rand_choice(def->size.x);

	if (hasItem(def, x, 10))
	{
	    continue;
	}
	addItem(def, x, 10, ITEM_ALTAR);
	naltar--;
    }

    // Add our links
    addLink(def, 0, 11, 100);
    addLink(def, def->size.x-1, 11, 100);
    addLink(def, 25, 24, 100);

    // Wire up caverns.
    ROOM_DEF		*cavern;

    def->south = "cavern1";
    
    cavern = build(BUILD_ROOMS, 1, "cavern1", 0);
    cavern->north = "world";
    cavern->south = "cavern2";
    cavern->east = "cavern1";
    cavern->west = "cavern1";
    registerDef((ROOM_DEF_NC*)cavern);

    cavern = build(BUILD_ROOMS, 2, "cavern2", 0);
    cavern->north = "cavern1";
    cavern->south = "cavern3";
    registerDef((ROOM_DEF_NC*)cavern);

    cavern = build(BUILD_ROOMS, 3, "cavern3", 0);
    cavern->north = "cavern2";
    cavern->south = "cavern4";
    registerDef((ROOM_DEF_NC*)cavern);

    cavern = build(BUILD_ROOMS, 4, "cavern4", 0);
    cavern->north = "cavern3";
    cavern->south = "cavern4";
    registerDef((ROOM_DEF_NC*)cavern);

    return def;
}

bool
BUILD::connectRooms(ROOM_DEF_NC *def, const RCT &a, const RCT &b)
{
    const SQUARE_NAMES		corridor = SQUARE_GROUND;
    const SQUARE_NAMES		wall = SQUARE_WALL;
    const SQUARE_NAMES		floor = SQUARE_FLOOR;
    const SQUARE_NAMES		empty = SQUARE_SOLIDEARTH;

    // Find our starting location
    IVEC2		p_a, p_b;

    PTRLIST<IVEC2>	valid, innervalid;
    for (int x = 2; x < a.w-2; x++)
    {
	valid.append(IVEC2(a.x+x, a.y));
	innervalid.append(IVEC2(a.x+x, a.y+1));
	valid.append(IVEC2(a.x+x, a.y+a.h-1));
	innervalid.append(IVEC2(a.x+x, a.y+a.h-2));
    }
    for (int y = 2; y < a.h-2; y++)
    {
	valid.append(IVEC2(a.x, a.y+y));
	innervalid.append(IVEC2(a.x+1, a.y+y));
	valid.append(IVEC2(a.x+a.w-1, a.y+y));
	innervalid.append(IVEC2(a.x+a.w-2, a.y+y));
    }

    int choice;
    choice = rand_choice(valid.entries());
    p_a = valid(choice);
    set(def, p_a, corridor);
    set(def, innervalid(choice), SQUARE_ALMOSTFLOOR);

    valid.clear();
    innervalid.clear();
    for (int x = 2; x < b.w-2; x++)
    {
	valid.append(IVEC2(b.x+x, b.y));
	innervalid.append(IVEC2(b.x+x, b.y+1));
	valid.append(IVEC2(b.x+x, b.y+b.h-1));
	innervalid.append(IVEC2(b.x+x, b.y+b.h-2));
    }
    for (int y = 2; y < b.h-2; y++)
    {
	valid.append(IVEC2(b.x, b.y+y));
	innervalid.append(IVEC2(b.x+1, b.y+y));
	valid.append(IVEC2(b.x+b.w-1, b.y+y));
	innervalid.append(IVEC2(b.x+b.w-2, b.y+y));
    }

    choice = rand_choice(valid.entries());
    p_b = valid(choice);
    set(def, p_b, corridor);
    set(def, innervalid(choice), SQUARE_ALMOSTFLOOR);

    return connectWithCorridor(def, p_a, p_b);
}

bool
BUILD::connectWithCorridor(ROOM_DEF_NC *def, IVEC2 p_a, IVEC2 p_b)
{
    ROOMDATA<int>		dist(def->width(), def->height(), -1);
    PTRLIST<IVEC2>		valid;

    const SQUARE_NAMES		corridor = SQUARE_GROUND;
    const SQUARE_NAMES		wall = SQUARE_WALL;
    const SQUARE_NAMES		empty = SQUARE_SOLIDEARTH;

    set(def, p_a, corridor);
    set(def, p_b, corridor);

    // Buid a distance map seeded at p_b
    valid.clear();
    valid.append(p_b);

    dist(p_b) = 0;

    while (valid.entries())
    {
	IVEC2		p = valid.removeFirst();
	int		d = dist(p);
	assert(d >= 0);

	// Update four way neighbours.  By doing this we encourage
	// orthogonal lines.
	int		nd = d + 2 + rand_choice(4);

	for (int angle = 0; angle < 4; angle++)
	{
	    int		dx, dy;
	    rand_getdirection(angle, dx, dy);
	    IVEC2	np = p;
	    np.x() += dx;
	    np.y() += dy;

	    if (dist.inbounds(np))
	    {
		// Check walkable!
		int nnd = nd;
		// Cheaper to re-use ,
		if (get(def, np) == corridor)
		    nnd--;

		if (get(def, np) == corridor ||
		    get(def, np) == empty)
		{
		    if (dist(np) < 0 || dist(np) > nnd)
		    {
			dist(np) = nnd;
			valid.append(np);
		    }
		}
	    }
	}
    }

    if (dist(p_a) < 0)
    {
	// Failed to connect!
	return false;
    }

    // Starting with p_a we have max steps given by dist(p_a)
    int		maxdist = (dist(p_a) + 2) * 2;
    int		step = 0;
    IVEC2	p = p_a;
    int		dx = 0, dy = 1;

    while (step < maxdist)
    {
	int		curd = dist(p);

	if (p == p_b)
	    break;

	// try last direction.
	IVEC2		np = p;
	np.x() += dx;	np.y() += dy;
	int		nd = dist.inbounds(np) ? dist(np) : -1;

	// 4/5 times try last direction to keep straightish.
	if (rand_choice(5) && (nd >= 0 && nd < curd))
	{
	    if (nd)
	    {
		// Replace
		set(def, np, corridor);
	    }
	    step++;
	    p = np;
	    continue;
	}

	// Look for a new downhill.
	int		angle = rand_choice(4);
	bool		found = false;
	for (int dir = 0; dir < 4; dir++)
	{
	    rand_getdirection(dir+angle, dx, dy);
	    np = p;
	    np.x() += dx;	np.y() += dy;
	    int		nd = dist.inbounds(np) ? dist(np) : -1;
	    if (nd >= 0 && nd < curd)
	    {
		if (nd)
		{
		    // Replace
		    set(def, np, corridor);
		}
		step++;
		p = np;
		found = true;
		break;
	    }
	}

	if (!found)
	{
	    // Nothing downwind???
	    return false;
	}
    }

    return true;
}

RCT
BUILD::rand_rect(ROOM_DEF_NC *def, int minw, int minh, int maxw, int maxh)
{
    RCT		result;

    assert(minw < def->width()-2);
    assert(minh < def->height()-2);

    result.w = rand_range(minw, maxw);
    result.h = rand_range(minh, maxh);

    result.w = MIN(result.w, def->width()-2);
    result.h = MIN(result.h, def->height()-2);

    result.x = rand_range(1, def->width()-result.w-1);
    result.y = rand_range(1, def->height()-result.h-1);

    return result;
}

ROOM_DEF_NC *
BUILD::buildRoomsAndCorridors(int w, int h, int level)
{
    ROOM_DEF_NC		*def = allocRoom(w, h);

    int		minrooms = 4;
    int		maxrooms = 6;

    def->level = level;

    while (1)
    {
	fill(def, SQUARE_SOLIDEARTH);

	PTRLIST<RCT> 	roomlist;

	// Try to place rooms until failures accumulate.
	while (1)
	{
	    if (roomlist.entries() >= maxrooms)
		break;

	    int 		fails = 0;
	    int			ntry = 50;
	    RCT		rect;
	    
	    // Pick the room size, then the position, so we
	    // don't bias to small rooms.
	    rect.w = rand_range(4+5, 5+5) + rand_choice(4) + rand_choice(3);
	    rect.h = rand_range(4+5, 5+5) + rand_choice(4) + rand_choice(3);

	    if (!roomlist.entries())
	    {
		rect.w = 9+5;
		rect.h = 9+5;
	    }
	    
	    while (fails < ntry)
	    {
		// We need walls and a single border on the + side
		// to ensure no two rooms abutt.
		// Our min size is a 2x2

		rect.x = rand_range(1, def->width()-rect.w-1);
		rect.y = rand_range(1, def->height()-rect.h-1);

		// See if we overlap
		bool		overlap = false;
		for (int i = 0; i < roomlist.entries(); i++)
		    if (rect.overlap(roomlist(i)))
		    {
			overlap = true;
			break;
		    }

		if (overlap)
		{
		    fails++;
		}
		else
		{
		    roomlist.append(rect);
		    break;
		}
	    }
	    if (fails == ntry)
		break;
	}

	// If too few rooms, quit.
	if (roomlist.entries() < minrooms)
	    continue;

	RCT		mainroom = roomlist(0);
	RCT		book1 = roomlist(1);
	RCT		book2 = roomlist(2);
	RCT		book3 = roomlist(3);

	// Stamp all the rooms.
	for (int i = 0; i < roomlist.entries(); i++)
	{
	    rect(def, roomlist(i).x, roomlist(i).y, roomlist(i).w-1, roomlist(i).h-1,
		    SQUARE_SKY);
	    rect(def, roomlist(i).x+1, roomlist(i).y+1, roomlist(i).w-3, roomlist(i).h-3,
		    SQUARE_WALL);
	    rect(def, roomlist(i).x+2, roomlist(i).y+2, roomlist(i).w-5, roomlist(i).h-5,
		    SQUARE_FLOOR);
	}

	// Ensure we are unbiased.
	roomlist.shuffle();

	// Connect all pairs
	bool		ok = true;
	for (int i = 1; i < roomlist.entries(); i++)
	{
	    RCT		a = roomlist(rand_choice(i));
	    RCT		b = roomlist(i);
	    // Remove padding
	    a.w--; a.h--;
	    b.w--; b.h--;
	    ok &= connectRooms(def, a, b);
	}

	// Connect to world above.
	{
	    IVEC2		src;
	    if (!randomSquare(def, SQUARE_GROUND, src.x(), src.y()))
		continue;
	    if (!connectWithCorridor(def, src, IVEC2(def->size.x/2, 0)))
		continue;
	    addLink(def, def->size.x/2, 0, 100);
	}
	// And below...
	if (level < 4)
	{
	    IVEC2		src;
	    if (!randomSquare(def, SQUARE_GROUND, src.x(), src.y()))
		continue;
	    if (!connectWithCorridor(def, src, IVEC2(def->size.x/2, def->size.y-1)))
		continue;
	    addLink(def, def->size.x/2, def->size.y-1, 100);
	}
	addLink(def, 0, def->size.y/2, 100);
	addLink(def, def->size.x-1, def->size.y/2, 100);

	// Replace sky with earth.
	replace(def, SQUARE_SKY, SQUARE_SOLIDEARTH, 100);

	// Failed to connect?
	if (!ok)
	    continue;

	addItem(def, mainroom.midx(), mainroom.midy(), ITEM_DAYSTAR);

	addItem(def, book1.midx(), book1.midy(), randomBook());
	addItem(def, book2.midx(), book2.midy(), randomBook());
	addItem(def, book3.midx(), book3.midy(), randomBook());

	// Add items & monsters now we know it is safe.
	// These are likely to end up in the main room.

	MOB_NAMES	mobperlevel[5] =
	{
	    MOB_TRILOBITE,
	    MOB_TRILOBITE, // 1
	    MOB_NAUTILUS,
	    MOB_CUBEMOB,
	    MOB_FINALMOB
	};
	for (int i = 0; i < 10; i++)
	{
	    int		x, y;

	    randomSquare(def, SQUARE_FLOOR, x, y);
	    addMob(def, x, y, mobperlevel[level]);
	}

	break;
    }

    return def;
}

void
BUILD::set(ROOM_DEF_NC *def, int x, int y, SQUARE_NAMES square)
{
    assert(x >= 0 && y >= 0);
    assert(x < def->size.x && y < def->size.y);

    def->squarelist[x + y * def->size.x] = square;
}

SQUARE_NAMES
BUILD::get(const ROOM_DEF_NC *def, int x, int y)
{
    if (x < 0 || y < 0 || x >= def->size.x || y >= def->size.y)
    {
	return SQUARE_NONE;
    }

    return def->squarelist[x + y * def->size.x];
}

void
BUILD::fill(ROOM_DEF_NC *def, SQUARE_NAMES square)
{
    rect(def, 0, 0, def->size.x, def->size.y, square);
}

void
BUILD::rect(ROOM_DEF_NC *def, int sx, int sy, int w, int h, SQUARE_NAMES square)
{
    for (int y = 0; y < h; y++)
    {
	if (y + sy < 0 || y + sy >= def->size.y)
	    continue;
	for (int x = 0; x < w; x++)
	{
	    if (x + sx < 0 || x + sx >= def->size.x)
		continue;
	    set(def, x+sx, y+sy, square);
	}
    }
}

ROOM_DEF_NC *
BUILD::allocRoom(int w, int h)
{
    ROOM_DEF_NC		*def = new ROOM_DEF_NC;

    def->size.x = w;
    def->size.y = h;
    def->squarelist = new SQUARE_NAMES[w * h];

    def->squareflaglist = allocPtList();
    def->itemlist = allocPtList();
    def->moblist = allocPtList();
    def->itemtypelist = allocPtList();
    def->moblevellist = allocPtList();
    def->intrinsiclist = allocPtList();
    def->signpostlist = allocPtList();
    def->linklist = allocPtList(-1);

    def->level = 0;
    def->placeholder = 0;
    def->rarity = 100;
    def->globalmapflags = 0;
    def->onlyonce = 0;
    def->allowstairs = 0;
    def->allowgeneration = 0;
    def->allowdig = 1;
    def->allowitems = 0;
    def->allowtel = 1;
    def->x = -1;
    def->y = -1;

    def->quest = 0;
    def->north = def->east = def->south = def->west = def->up = def->down = 0;
    def->name = 0;
    
    fill(def, SQUARE_NONE);

    return def;
}

PT2 *
BUILD::allocPtList(int lastdata)
{
    PT2		*list = new PT2[1];
    list->x = list->y = -1;
    list->data = lastdata;

    return list;
}

PT2 *
BUILD::appendPtList(PT2 *oldlist, PT2 newpt)
{
    int		 len;
    PT2		*list;

    for (len = 0; oldlist[len].data > 0; len++);
    len++;		// include null


    list = new PT2[len+1];
    memcpy(list, oldlist, len * sizeof(PT2));
    delete [] oldlist;
    // Copy forward null
    list[len] = list[len-1];
    // And insert our new point
    list[len-1] = newpt;

    return list;
}

static int
findClass(ROOMDATA<int> &fclass, int x, int y)
{
    int		c = fclass(x, y);
    if (c < 0)
	return c;

    int		linidx = x + y * fclass.width();
    if (c == linidx)
	return c;

    int		nx, ny;
    nx = c % fclass.width();
    ny = c / fclass.width();

    c = findClass(fclass, nx, ny);
    fclass(x, y) = c;

    return c;
}

void
unionSquares(ROOMDATA<int> &fclass, int x1, int y1, int x2, int y2)
{
    int		c1 = findClass(fclass, x1, y1);
    int		c2 = findClass(fclass, x2, y2);
    // We union the new tree with the old tree, 
    // this then preserves the rank order.
    if (c1 == c2)
	return;
    if (c1 < c2)
    {
	int		nx, ny;
	nx = c2 % fclass.width();
	ny = c2 / fclass.width();
	fclass(nx, ny) = c1;
    }
    else
    {
	int		nx, ny;
	nx = c1 % fclass.width();
	ny = c1 / fclass.width();
	fclass(nx, ny) = c2;
    }
}

void
BUILD::findConnected(ROOM_DEF_NC *def, bool allowwater, bool allowfire,
		ROOMDATA<int> &connect, PTRLIST<int> &count)
{
    int			w = def->size.x, h = def->size.y;
    ROOMDATA<int>	fclass(w, h, -1);

    for (int y = 0; y < h; y++)
    {
	for (int x = 0; x < w; x++)
	{
	    // Illegal tiles get -1
	    SQUARE_NAMES	square = get(def, x, y);
	    if (glb_squaredefs[square].connectable)
	    {
		// See if we match the water and fire.
		if ( (!glb_squaredefs[square].connectable_fire || allowfire) &&
		     (!glb_squaredefs[square].connectable_water || allowwater) )
		{
		    fclass(x, y) = x + y * w;
		}
	    }
	}
    }

    // Two way connectivity.
    for (int y = 0; y < h; y++)
    {
	for (int x = 0; x < w; x++)
	{
	    if (findClass(fclass, x, y) < 0)
		continue;

	    if (x > 0 && findClass(fclass, x-1, y) >= 0)
	    {
		// Connect back.
		unionSquares(fclass, x-1, y, x, y);
	    }
	    if (y > 0 && findClass(fclass, x, y-1) >= 0)
	    {
		unionSquares(fclass, x, y-1, x, y);
	    }
	}
    }

    // Create our lookup table to remap our classes.
    ROOMDATA<int>	lut(w, h, -1);
    int			nclass = 0;
    for (int y = 0; y < h; y++)
    {
	for (int x = 0; x < w; x++)
	{
	    int c = findClass(fclass, x, y);

	    if (c >= 0)
	    {
		// See if this class has a lut entry yet.
		int nx = c % w;
		int ny = c / w;
		if (lut(nx, ny) == -1)
		{
		    lut(nx, ny) = nclass++;
		}
	    }
	}
    }

    // Now remap all the classes and count!
    count.resize(0);
    count.resize(nclass);

    for (int y = 0; y < h; y++)
    {
	for (int x = 0; x < w; x++)
	{
	    int c = findClass(fclass, x, y);

	    if (c >= 0)
	    {
		int nx = c % w;
		int ny = c / w;

		int	value = lut(nx, ny);
		count(value)++;
		connect(x, y) = value;
	    }
	    else
		connect(x, y) = -1;
	}
    }
}

void
BUILD::addLink(ROOM_DEF_NC *def, int x, int y, int link)
{
    PT2		pt;
    pt.x = x;
    pt.y = y;
    pt.data = link;

    def->linklist = appendPtList(def->linklist, pt);
}

bool
BUILD::hasItem(ROOM_DEF_NC *def, int x, int y)
{
    for (int i = 0; def->itemlist[i].data; i++)
    {
	if (def->itemlist[i].x == x && def->itemlist[i].y == y)
	    return true;
    }

    return false;
}

void
BUILD::addItem(ROOM_DEF_NC *def, int x, int y, ITEM_NAMES item)
{
    PT2		pt;
    pt.x = x;
    pt.y = y;
    pt.data = item;

    def->itemlist = appendPtList(def->itemlist, pt);
}

void
BUILD::addMob(ROOM_DEF_NC *def, int x, int y, MOB_NAMES mob)
{
    PT2		pt;
    pt.x = x;
    pt.y = y;
    pt.data = mob;

    def->moblist = appendPtList(def->moblist, pt);
}

void
BUILD::replace(ROOM_DEF_NC *def,
		SQUARE_NAMES oldsquare,
		SQUARE_NAMES newsquare,
		int chance)
{
    rectReplace(def, 0, 0, def->size.x, def->size.y,
		    oldsquare, newsquare, chance);
}

void
BUILD::rectReplace(ROOM_DEF_NC *def,
		int sx, int sy, int w, int h,
		SQUARE_NAMES oldsquare,
		SQUARE_NAMES newsquare,
		int chance)
{
    for (int y = 0; y < h; y++)
    {
	if (y + sy < 0 || y + sy >= def->size.y)
	    continue;
	for (int x = 0; x < w; x++)
	{
	    if (x + sx < 0 || x + sx >= def->size.x)
		continue;
	    if (get(def, x+sx, y+sy) == oldsquare)
	    {
		if (rand_chance(chance))
		    set(def, x+sx, y+sy, newsquare);
	    }
	}
    }
}

bool
BUILD::findRandomMazePassage(ROOM_DEF_NC *def, SQUARE_NAMES floor, int &rnx, int &rny, int &rdx, int &rdy)
{
    int			nvalid = 0;

    for (int y = 1; y < def->size.y; y += 2)
    {
	for (int x = 1; x < def->size.x; x += 2)
	{
	    if (get(def, x, y) == floor)
		continue;

	    int		dx, dy;

	    FORALL_4DIR(dx, dy)
	    {
		if (get(def, x + dx*2, y + dy*2) == floor)
		{
		    nvalid++;
		    if (!rand_choice(nvalid))
		    {
			rnx = x;		rny = y;
			rdx = dx;		rdy = dy;
		    }
		}
	    }
	}
    }

    return nvalid > 0;
}

int
BUILD::numExits(ROOM_DEF_NC *def, SQUARE_NAMES floor, int x, int y) 
{
    int		nexit = 0;
    int		dx, dy;

    FORALL_4DIR(dx, dy)
    {
	if (get(def, x+dx, y+dy) == floor)
	    nexit++;
    }
    return nexit;
}


void
BUILD::tryBlockCorridor(ROOM_DEF_NC *def, int x, int y, int chance, SQUARE_NAMES floor, SQUARE_NAMES blocker)
{
    if (get(def, x, y) == floor)
    {
	// Could be a corridor..
	// Either vertical are free or horizontal are free.
	bool		floordir[4];
	int		dx, dy;
	FORALL_4DIR(dx, dy)
	{
	    floordir[lcl_angle] = (get(def, x+dx, y+dy) == floor);
	}

	if ( (floordir[0] && floordir[2] && !floordir[1] && !floordir[3]) ||
	     (!floordir[0] && !floordir[2] && floordir[1] && floordir[3]))
	{
	    // Corridor!
	    if (rand_chance(chance))
		set(def, x, y, blocker);
	}
    }
}

void
BUILD::cavern(ROOM_DEF_NC *def, int percentfloor, SQUARE_NAMES floor, SQUARE_NAMES wall)
{
    set(def, def->size.x/2, def->size.y/2, floor);
    cavern(def, percentfloor, floor, wall, 
	    def->size.x/2, def->size.y/2);
}

void
BUILD::cavern(ROOM_DEF_NC *def, int percentfloor, SQUARE_NAMES floor, SQUARE_NAMES wall, int centerx, int centery)
{
    int		totaltoeat = (def->size.x-2) * (def->size.y-2) * percentfloor / 100;
    int		eaten = 0;

    fill(def, wall);
    PTRLIST<PT2>	starts;

    set(def, centerx, centery, floor);
    eaten++;
    starts.append(point(centerx, centery));

    while (eaten < totaltoeat)
    {
	int		ex, ey;
	int		choice;

	choice = rand_choice(starts.entries());
	ex = starts(choice).x;
	ey = starts(choice).y;

	int		len = rand_range(10, 20);
	int		dir = rand_range(0, 3);
	while (len--)
	{
	    int		dx, dy;
	    rand_getdirection(dir, dx, dy);
	    ex += dx;
	    ey += dy;
	    if (ex < 1 || ex >= def->size.x-1)
		break;
	    if (ey < 1 || ey >= def->size.y-1)
		break;

	    if (get(def, ex, ey) == wall)
	    {
		set(def, ex, ey, floor);
		starts.append(point(ex, ey));
		eaten++;
	    }

	    // Change direction.
	    // Do not backtrack!
	    dir += rand_choice(3) - 1;
	    dir &= 3;
	}
    }

}

void
BUILD::blockNodes(ROOM_DEF_NC *def, int chance, SQUARE_NAMES floor, SQUARE_NAMES blocker)
{
    // Assuming a maze, run over all nodes and block 
    // sometimes.
    for (int y = 1; y < def->size.y; y += 2)
    {
	for (int x = 1; x < def->size.x; x += 2)
	{
	    if (get(def, x, y) == floor)
	    {
		if (rand_chance(chance))
		    set(def, x, y, blocker);
	    }
	}
    }
}

void
BUILD::blockCorridors(ROOM_DEF_NC *def, int chance, SQUARE_NAMES floor, SQUARE_NAMES blocker)
{
    // Assuming a maze, run over all corridors and block straight
    // lines... sometimes.
    for (int y = 2; y < def->size.y; y += 2)
    {
	for (int x = 2; x < def->size.x; x += 2)
	{
	    tryBlockCorridor(def, x-1, y, chance, floor, blocker);
	    tryBlockCorridor(def, x, y-1, chance, floor, blocker);
	}
    }
    // SPECIAL CASE: Exit is always blocked!
    for (int y = 0; y < def->size.y; y++)
    {
	int x = def->size.x-1;
	if (get(def, x, y) == floor)
	    set(def, x, y, blocker);
    }
}

void
BUILD::maze(ROOM_DEF_NC *def, SQUARE_NAMES floor, SQUARE_NAMES wall)
{
    fill(def, SQUARE_WALL);
    rect(def, 1, 1, def->size.x-2, def->size.y-2, wall);

    int		sx, sy;

    sx = 0;
    sy = rand_range(0, def->size.y/2 - 1);
    sy = sy * 2 + 1;

    set(def, sx, sy, floor);
    addLink(def, sx, sy, 100);

    set(def, sx+1, sy, floor);

    sx = def->size.x-1;
    sy = rand_range(0, def->size.y/2 - 1);
    sy = sy * 2 + 1;
    set(def, sx, sy, floor);
    addLink(def, sx, sy, 100);
    set(def, sx-1, sy, floor);

    // Start with disconnected cells.
    int			nx, ny, dx, dy;
    while (findRandomMazePassage(def, floor, nx, ny, dx, dy))
    {
	set(def, nx, ny, floor);
	set(def, nx+dx, ny+dy, floor);
    }

    // Now open all the dead ends.
    for (int y = 1; y < def->size.y; y += 2)
    {
	for (int x = 1; x < def->size.x; x += 2)
	{
	    if (numExits(def, floor, x, y) == 1)
	    {
		int		nescape = 0, edx, edy;
		FORALL_4DIR(dx, dy)
		{
		    if (get(def, x + dx*2, y + dy*2) == floor)
		    {
			nescape++;
			if (!rand_choice(nescape))
			{
			    edx = dx;
			    edy = dy;
			}
		    }
		}
		if (nescape)
		{
		    set(def, x+edx, y+edy, floor);
		}
	    }
	}
    }

}

void
BUILD::registerDef(ROOM_DEF_NC *def)
{
    MAP		*map = new MAP(* (ROOM_DEF *)def);
    glbAtlas()->registerMap(map);
    glbAtlas()->registerDef(def);
}

int
BUILD::countSquare(ROOM_DEF_NC *def, SQUARE_NAMES square)
{
    int		count = 0;
    for (int y = 0; y < def->size.y; y++)
	for (int x = 0; x < def->size.x; x++)
	{
	    if (get(def, x, y) == square)
	    {
		count++;
	    }
	}
    return count;
}

bool
BUILD::nthSquare(ROOM_DEF_NC *def, int nth, SQUARE_NAMES square, int &x, int &y)
{
    for (y = 0; y < def->size.y; y++)
	for (x = 0; x < def->size.x; x++)
	{
	    if (get(def, x, y) == square)
	    {
		if (!nth)
		    return true;
		nth--;
	    }
	}
    return false;
}

bool
BUILD::randomSquare(ROOM_DEF_NC *def, SQUARE_NAMES square, int &x, int &y)
{
    int		num = countSquare(def, square);

    x = y = -1;
    if (!num)
	return false;

    return nthSquare(def, rand_choice(num), square, x, y);
}

int
BUILD::randomSquareInColumn(ROOM_DEF_NC *def, SQUARE_NAMES square, int col)
{
    int		result = -1;
    int		total = 0;

    for (int y = 0; y < def->size.y; y++)
    {
	if (get(def, col, y) == square)
	{
	    total++;
	    if (!rand_choice(total))
		result = y;
	}
    }
    return result;
}

void
BUILD::shrinkToFit(ROOM_DEF_NC *def, int xpad, int ypad, SQUARE_NAMES wall)
{
    int		left, right;
    int		top, bot;
    int		x, y;

    for (left = 0; left < def->size.x; left++)
    {
	for (y = 0; y < def->size.y; y++)
	{
	    if (get(def, left, y) != wall)
		break;
	}
	if (y != def->size.y)
	    break;
    }

    for (right = def->size.x; right --> 0; )
    {
	for (y = 0; y < def->size.y; y++)
	{
	    if (get(def, right, y) != wall)
		break;
	}
	if (y != def->size.y)
	    break;
    }

    for (top = 0; top < def->size.y; top++)
    {
	for (x = 0; x < def->size.x; x++)
	{
	    if (get(def, x, top) != wall)
		break;
	}
	if (x != def->size.x)
	    break;
    }

    for (bot = def->size.y; bot --> 0; )
    {
	for (x = 0; x < def->size.x; x++)
	{
	    if (get(def, x, bot) != wall)
		break;
	}
	if (x != def->size.x)
	    break;
    }

    int		ow = def->size.x;
    int		oh = def->size.y;
    SQUARE_NAMES	*os = def->squarelist;

    def->size.x = (right - left) + 1 + 2*xpad;
    def->size.y = (bot - top) + 1 + 2*ypad;

    def->squarelist = new SQUARE_NAMES[def->size.x * def->size.y];

    fill(def, wall);

    // Now copy in center!
    for (y = 0; y < bot-top+1; y++)
    {
	for (x = 0; x < right-left+1; x++)
	{
	    set(def, x+xpad, y+ypad, 
		    os[ (x + left) + (y + top) * ow ]);
	}
    }

    delete [] os;
}

char *
BUILD::createUniqueName(const char *prefix)
{
    static int suffix = 0;

    BUF		buf;
    buf.sprintf("%s_%d", prefix, suffix);
    suffix++;

    return strdup(buf.buffer());
}

static int
findClass(PTRLIST<int> &connect, int idx)
{
    int		c = connect(idx);

    if (c < 0)
	return c;

    if (c == idx)
	return c;

    c = findClass(connect, c);
    connect(idx) = c;

    return c;
}

static void
unionClass(PTRLIST<int> &connect, int aidx, int bidx)
{
    int ca = findClass(connect, aidx);
    int cb = findClass(connect, bidx);
    if (ca == cb)
	return;

    if (ca < cb)
	connect(cb) = ca;
    else
	connect(ca) = cb;
    findClass(connect, aidx);
    findClass(connect, bidx);
    assert(findClass(connect, aidx) == findClass(connect, bidx));
}

bool
BUILD::drawRiver(ROOM_DEF_NC *def,
			const ROOMDATA<float> &height,
			int x, int y,
			SQUARE_NAMES water,
			SQUARE_NAMES dest)
{
    int		dx, dy;
    int		cdx, cdy;
    float	dist = height(x, y);

    cdx = 1;
    cdy = 0;
    while (get(def, x, y) != dest)
    {
	bool		found = false;
#if 1
	float	ndist;
	int	choice = 0;
	FORALL_4DIR(dx, dy)
	{
	    if (height(x+dx, y+dy) > dist)
	    {
		choice++;
		if (!rand_choice(choice))
		{
		    cdx = dx;
		    cdy = dy;
		    ndist = height(x+dx, y+dy);
		}
	    }
	}
	if (choice)
	{
	    found = true;
	    dist = ndist;
	}
#else
	// Steepest descent.
	FORALL_4DIR(dx, dy)
	{
	    if (height(x+dx, y+dy) > dist)
	    {
		found = true;
		dist = height(x+dx, y+dy);
		cdx = dx;
		cdy = dy;
	    }
	}
#endif
	x += cdx;
	y += cdy;
	if (get(def, x, y) == dest)
	    break;

	if (!found)
	{
	    // Gah!
	    // Make a pool, ensure we are not in mountains
	    rect(def, x-1, y-1, 3, 3, water);

	    // FInd closest destination and draw a line right there.
	    int		tx = x, ty = y;
	    findClosest(def, dest, tx, ty);

	    assert(get(def, tx, ty) == dest);

	    while (x != tx || y != ty)
	    {
		int		dx = SIGN(tx-x);
		int		dy = SIGN(ty-y);
		x += dx;
		set(def, x, y, water);
		y += dy;
		set(def, x, y, water);
	    }

	    return true;
	}

	set(def, x, y, water);
    }
    return true;
}

void
BUILD::findClosest(ROOM_DEF_NC *def, SQUARE_NAMES goal,
		    int &sx, int &sy)
{
    int		r = 0;

    for (r = 0; r < MAX(def->size.x, def->size.y); r++)
    {
	for (int sr = 0; sr <= r; sr++)
	{
	    int		 x = sx;
	    int		 y = sy;
	    if (get(def, x-r, y-sr) == goal)
	    {
		sx = x-r;
		sy = y-sr;
		return;
	    }
	    if (get(def, x-r, y+sr) == goal)
	    {
		sx = x-r;
		sy = y+sr;
		return;
	    }
	    if (get(def, x+r, y-sr) == goal)
	    {
		sx = x+r;
		sy = y-sr;
		return;
	    }
	    if (get(def, x+r, y+sr) == goal)
	    {
		sx = x+r;
		sy = y+sr;
		return;
	    }
	    if (get(def, x-sr, y-r) == goal)
	    {
		sx = x-sr;
		sy = y-r;
		return;
	    }
	    if (get(def, x-sr, y+r) == goal)
	    {
		sx = x-sr;
		sy = y+r;
		return;
	    }
	    if (get(def, x+sr, y-r) == goal)
	    {
		sx = x+sr;
		sy = y-r;
		return;
	    }
	    if (get(def, x+sr, y+r) == goal)
	    {
		sx = x+sr;
		sy = y+r;
		return;
	    }
	}
    }
}

void
BUILD::findFarthest(ROOM_DEF_NC *def, SQUARE_NAMES goal,
		    int &rsx, int &rsy)
{
    int		r = 0;
    int		sx = rsx, sy = rsy;

    for (r = 0; r < MAX(def->size.x, def->size.y); r++)
    {
	for (int sr = 0; sr <= r; sr++)
	{
	    int		 x = sx;
	    int		 y = sy;
	    if (get(def, x-r, y-sr) == goal)
	    {
		rsx = x-r;
		rsy = y-sr;
	    }
	    if (get(def, x-r, y+sr) == goal)
	    {
		rsx = x-r;
		rsy = y+sr;
	    }
	    if (get(def, x+r, y-sr) == goal)
	    {
		rsx = x+r;
		rsy = y-sr;
	    }
	    if (get(def, x+r, y+sr) == goal)
	    {
		rsx = x+r;
		rsy = y+sr;
	    }
	    if (get(def, x-sr, y-r) == goal)
	    {
		rsx = x-sr;
		rsy = y-r;
	    }
	    if (get(def, x-sr, y+r) == goal)
	    {
		rsx = x-sr;
		rsy = y+r;
	    }
	    if (get(def, x+sr, y-r) == goal)
	    {
		rsx = x+sr;
		rsy = y-r;
	    }
	    if (get(def, x+sr, y+r) == goal)
	    {
		rsx = x+sr;
		rsy = y+r;
	    }
	}
    }
}

void
BUILD::buildTruePermutation(PTRLIST<int> &list, int len)
{
    list.resize(0);
    for (int i = 0; i < len; i++)
	list.append(i);
    list.shuffle();
    // Small lists can't be fixed.
    if (list.entries() < 2)
	return;
    // We may have self-references.
    // RUn through everything.  If we find an illegal, swap with
    // someone who won't be unhappy coming into our spot
    for (int i = 0; i < list.entries(); i++)
    {
	if (list(i) == i)
	{
	    // Uh-oh!  Don't want to get ourself as a prize :>
	    // By construction, we can always swap ourselves with
	    // a different random entry since it must not equal
	    // us and we won't equal its rank!
	    int		j = rand_choice(list.entries()-1);
	    if (j >= i)
		j++;
	    list.swapEntries(i, j);
	}
    }
}
